"""Defines the metagrating challenge.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import functools
from typing import Any, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp
from fmmax import basis, fmm  # type: ignore[import-untyped]
from jax import tree_util
from totypes import symmetry, types

from invrs_gym.challenges import base
from invrs_gym.challenges.diffract import common
from invrs_gym.utils import initializers

AVERAGE_EFFICIENCY = "average_efficiency"
MIN_EFFICIENCY = "min_efficiency"

density_initializer = functools.partial(
    initializers.noisy_density_initializer,
    relative_mean=0.5,
    relative_noise_amplitude=0.1,
)


class MetagratingComponent(base.Component):
    """Defines a metagrating component."""

    def __init__(
        self,
        spec: common.GratingSpec,
        sim_params: common.GratingSimParams,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs: Any,
    ) -> None:
        """Initializes the grating component.

        Args:
            spec: Defines the physical specification of the grating.
            sim_params: Defines simulation parameters for the grating.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """

        self.spec = spec
        self.sim_params = sim_params
        self.seed_density = common.seed_density(
            self.sim_params.grid_shape, **seed_density_kwargs
        )
        self.density_initializer = density_initializer

        self.expansion = basis.generate_expansion(
            primitive_lattice_vectors=basis.LatticeVectors(
                u=self.spec.period_x * basis.X,
                v=self.spec.period_y * basis.Y,
            ),
            approximate_num_terms=self.sim_params.approximate_num_terms,
            truncation=self.sim_params.truncation,
        )

    def init(self, key: jax.Array) -> types.Density2DArray:
        """Return the initial parameters for the metagrating component."""
        params = self.density_initializer(key, self.seed_density)
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: types.Density2DArray,
        *,
        wavelength: Optional[Union[float, jnp.ndarray]] = None,
        expansion: Optional[basis.Expansion] = None,
    ) -> Tuple[common.GratingResponse, base.AuxDict]:
        """Computes the response of the metagrating.

        Args:
            params: The parameters defining the metagrating, matching those returned
                by the `init` method.
            wavelength: Optional wavelength to override the default in `sim_params`.
            expansion: Optional expansion to override the default `expansion`.

        Returns:
            The `(response, aux)` tuple.
        """
        if expansion is None:
            expansion = self.expansion
        if wavelength is None:
            wavelength = self.sim_params.wavelength
        transmission_efficiency, reflection_efficiency = common.grating_efficiency(
            density_array=params.array,  # type: ignore[arg-type]
            spec=self.spec,
            wavelength=jnp.asarray(wavelength),
            polarization=self.sim_params.polarization,
            expansion=expansion,
            formulation=self.sim_params.formulation,
        )
        response = common.GratingResponse(
            wavelength=jnp.asarray(wavelength),
            transmission_efficiency=transmission_efficiency,
            reflection_efficiency=reflection_efficiency,
            expansion=expansion,
        )
        return response, {}


@dataclasses.dataclass
class MetagratingChallenge(base.Challenge):
    """Defines the metagrating challenge.

    The objective of the metagrating challenge is to design a density so that incident
    light is efficiently diffracted into the specified transmission order. The
    challenge is considered solved when the transmission is above the lower bound.

    Attributes:
        component: The component to be designed.
        transmission_order: The transmission diffraction order to be maximized.
        transmission_lower_bound: The lower bound for transmission into the specified
            order. When this value is exceeded, the challenge is considered solved.
    """

    component: MetagratingComponent
    transmission_order: Tuple[int, int]
    transmission_lower_bound: float

    def loss(self, response: common.GratingResponse) -> jnp.ndarray:
        """Compute a scalar loss from the component `response`."""
        efficiency = _value_for_order(
            response.transmission_efficiency,
            expansion=response.expansion,
            order=self.transmission_order,
        )
        return jnp.mean(1 - jnp.abs(jnp.sqrt(efficiency)))

    def distance_to_target(self, response: common.GratingResponse) -> jnp.ndarray:
        """Compute distance from the component `response` to the challenge target."""
        efficiency = _value_for_order(
            response.transmission_efficiency,
            expansion=response.expansion,
            order=self.transmission_order,
        )
        elementwise_distance_to_window = jnp.maximum(
            0, self.transmission_lower_bound - efficiency
        )
        return jnp.linalg.norm(elementwise_distance_to_window)

    def metrics(
        self,
        response: common.GratingResponse,
        params: types.Density2DArray,
        aux: base.AuxDict,
    ) -> base.AuxDict:
        """Compute challenge metrics."""
        del params, aux
        efficiency = _value_for_order(
            response.transmission_efficiency,
            expansion=response.expansion,
            order=self.transmission_order,
        )
        return {
            AVERAGE_EFFICIENCY: jnp.mean(efficiency),
            MIN_EFFICIENCY: jnp.amin(efficiency),
        }


def _value_for_order(
    array: jnp.ndarray,
    expansion: basis.Expansion,
    order: Tuple[int, int],
) -> jnp.ndarray:
    """Extracts the value from `array` for the specified Fourier order."""
    order_idx = common.index_for_order(order, expansion)
    return array[..., order_idx, :]


# -----------------------------------------------------------------------------
# Metagrating with 1.371 x 0.525 um design region.
# -----------------------------------------------------------------------------


METAGRATING_SPEC = common.GratingSpec(
    permittivity_ambient=(1.0 + 0.0j) ** 2,
    permittivity_grating=(3.45 + 0.00001j) ** 2,
    permittivity_encapsulation=(1.0 + 0.00001j) ** 2,
    permittivity_substrate=(1.45 + 0.0j) ** 2,
    thickness_grating=0.325,
    period_x=float(1.050 / jnp.sin(jnp.deg2rad(50.0))),
    period_y=0.525,
)

METAGRATING_SIM_PARAMS = common.GratingSimParams(
    grid_shape=(118, 45),
    wavelength=1.050,
    polarization=common.TM,
    formulation=fmm.Formulation.JONES_DIRECT,
    approximate_num_terms=200,
    truncation=basis.Truncation.CIRCULAR,
)

SYMMETRIES = (symmetry.REFLECTION_E_W,)

# Objective is to diffract light into the +1 transmitted order, with efficiency better
# than 95 percent.
TRANSMISSION_ORDER = (1, 0)
TRANSMISSION_LOWER_BOUND = 0.95


def metagrating(
    minimum_width: int = 7,
    minimum_spacing: int = 7,
    density_initializer: base.DensityInitializer = density_initializer,
    transmission_order: Tuple[int, int] = TRANSMISSION_ORDER,
    transmission_lower_bound: float = TRANSMISSION_LOWER_BOUND,
    spec: common.GratingSpec = METAGRATING_SPEC,
    sim_params: common.GratingSimParams = METAGRATING_SIM_PARAMS,
    symmetries: Sequence[str] = SYMMETRIES,
) -> MetagratingChallenge:
    """Metagrating challenge with 1.371 x 0.525 um design region.

    The metagrating challenge is based on the metagrating example in "Validation and
    characterization of algorithms for photonics inverse design" by Chen et al.
    (in preparation).

    It involves maximizing diffraction of light transmitted from a silicon oxide
    substrate into the ambient using a patterned silicon metastructure. The excitation
    is TM-polarized plane wave with 1.05 micron wavelength.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.
        transmission_order: The diffraction order to be maximized.
        transmission_lower_bound: The lower bound for transmission. When the lower
            bound is exceeded, the challenge is considered to be solved.
        spec: Defines the physical specification of the metagrating.
        sim_params: Defines the simulation settings of the metagrating.
        symmetries: Defines the symmetries of the metagrating.

    Returns:
        The `MetagratingChallenge`.
    """
    return MetagratingChallenge(
        component=MetagratingComponent(
            spec=spec,
            sim_params=sim_params,
            density_initializer=density_initializer,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            symmetries=symmetries,
        ),
        transmission_order=transmission_order,
        transmission_lower_bound=transmission_lower_bound,
    )



"""Defines defaults for the ceviche challenges.

Copyright (c) 2023 The INVRS-IO authors.
"""

from typing import Union

import jax.numpy as jnp
from ceviche_challenges import (  # type: ignore[import-untyped]
    beam_splitter,
    mode_converter,
    model_base,
    params,
)
from ceviche_challenges import units as u
from ceviche_challenges import waveguide_bend, wdm  # type: ignore[import-untyped]
from totypes import symmetry

DeviceSpec = Union[
    beam_splitter.spec.BeamSplitterSpec,
    mode_converter.spec.ModeConverterSpec,
    waveguide_bend.spec.WaveguideBendSpec,
    wdm.spec.WdmSpec,
]
Model = model_base.Model


def _linear_from_decibels(x_decibels: float) -> jnp.ndarray:
    """Converts a quantity in decibels to linear."""
    linear: jnp.ndarray = jnp.asarray(10 ** (x_decibels / 10))
    return linear


WG_WIDTH = 400 * u.nm
WG_MODE_PADDING = 520 * u.nm
WG_LENGTH = 720 * u.nm

PADDING = 400 * u.nm
PORT_PML_OFFSET = 40 * u.nm

CLADDING_PERMITTIVITY = 2.25
SLAB_PERMITTIVITY = 12.25
INPUT_MONITOR_OFFSET = 40 * u.nm

PML_WIDTH_GRIDPOINTS = 20


# -----------------------------------------------------------------------------
# Default simulation parameters for standard and lightweight challenges.
# -----------------------------------------------------------------------------


SIM_PARAMS = params.CevicheSimParams(
    resolution=10 * u.nm,
    wavelengths=u.Array([1265.0, 1270.0, 1275.0, 1285.0, 1290.0, 1295.0], u.nm),
)

LIGHTWEIGHT_SIM_PARAMS = params.CevicheSimParams(
    resolution=40 * u.nm,
    wavelengths=u.Array([1270.0, 1290.0], u.nm),
)


# -----------------------------------------------------------------------------
# Default length scale targets for standard and lightweight challenges.
# -----------------------------------------------------------------------------


MINIMUM_WIDTH = 8  # 80 nm at the standard 10 nm resolution.
MINIMUM_SPACING = 8
LIGHTWEIGHT_MINIMUM_WIDTH = 3  # 120 nm at the lightweight 40 nm resolution.
LIGHTWEIGHT_MINIMUM_SPACING = 3


# -----------------------------------------------------------------------------
# Beamsplitter with a 2.0 x 3.2 um design region.
# -----------------------------------------------------------------------------


_BEAM_SPLITTER_SPEC = beam_splitter.spec.BeamSplitterSpec(
    wg_width=WG_WIDTH,
    wg_length=WG_LENGTH,
    wg_separation=1040 * u.nm,
    wg_mode_padding=WG_MODE_PADDING,
    port_pml_offset=PORT_PML_OFFSET,
    variable_region_size=(3200 * u.nm, 2000 * u.nm),
    cladding_permittivity=CLADDING_PERMITTIVITY,
    slab_permittivity=SLAB_PERMITTIVITY,
    input_monitor_offset=INPUT_MONITOR_OFFSET,
    design_symmetry=None,
    pml_width=PML_WIDTH_GRIDPOINTS,
)

BEAM_SPLITTER_MODEL = beam_splitter.model.BeamSplitterModel(
    params=SIM_PARAMS, spec=_BEAM_SPLITTER_SPEC
)
LIGHTWEIGHT_BEAM_SPLITTER_MODEL = beam_splitter.model.BeamSplitterModel(
    params=LIGHTWEIGHT_SIM_PARAMS, spec=_BEAM_SPLITTER_SPEC
)
BEAM_SPLITTER_TRANSMISSION_LOWER_BOUND = jnp.asarray(
    [
        0.0,  # |S11|^2 lower bound
        _linear_from_decibels(-3.5),  # |S21|^2 lower bound
        _linear_from_decibels(-3.5),  # |S31|^2 lower bound
        0.0,  # |S41|^2 lower bound
    ]
)
BEAM_SPLITTER_TRANSMISSION_UPPER_BOUND = jnp.asarray(
    [
        _linear_from_decibels(-20.0),  # |S11|^2 upper bound
        1 - _linear_from_decibels(-3.5),  # |S21|^2 upper bound
        1 - _linear_from_decibels(-3.5),  # |S31|^2 upper bound
        _linear_from_decibels(-20),  # |S41|^2 upper bound
    ]
)
BEAM_SPLITTER_SYMMETRIES = (symmetry.REFLECTION_E_W, symmetry.REFLECTION_N_S)


# -----------------------------------------------------------------------------
# Mode converter with a 1.6 x 1.6 um design region.
# -----------------------------------------------------------------------------


_MODE_CONVERTER_SPEC = mode_converter.spec.ModeConverterSpec(
    left_wg_width=WG_WIDTH,
    left_wg_mode_padding=WG_MODE_PADDING,
    left_wg_mode_order=1,  # Fundamental mode.
    right_wg_width=WG_WIDTH,
    right_wg_mode_padding=WG_MODE_PADDING,
    right_wg_mode_order=2,  # Second mode.
    wg_length=WG_LENGTH,
    padding=PADDING,
    port_pml_offset=PORT_PML_OFFSET,
    variable_region_size=(1600 * u.nm, 1600 * u.nm),
    cladding_permittivity=CLADDING_PERMITTIVITY,
    slab_permittivity=SLAB_PERMITTIVITY,
    input_monitor_offset=INPUT_MONITOR_OFFSET,
    pml_width=PML_WIDTH_GRIDPOINTS,
)

MODE_CONVERTER_MODEL = mode_converter.model.ModeConverterModel(
    params=SIM_PARAMS, spec=_MODE_CONVERTER_SPEC
)
LIGHTWEIGHT_MODE_CONVERTER_MODEL = mode_converter.model.ModeConverterModel(
    params=LIGHTWEIGHT_SIM_PARAMS, spec=_MODE_CONVERTER_SPEC
)
MODE_CONVERTER_TRANSMISSION_LOWER_BOUND = jnp.asarray(
    [
        0.0,  # |S11|^2 lower bound
        _linear_from_decibels(-0.5),  # |S21|^2 lower bound
    ]
)
MODE_CONVERTER_TRANSMISSION_UPPER_BOUND = jnp.asarray(
    [
        _linear_from_decibels(-20.0),  # |S11|^2 upper bound
        1.0,  # |S21|^2 upper bound
    ]
)

# -----------------------------------------------------------------------------
# Waveguide bend with a 1.6 x 1.6 um design region.
# -----------------------------------------------------------------------------


_WAVEGUIDE_BEND_SPEC = waveguide_bend.spec.WaveguideBendSpec(
    wg_width=WG_WIDTH,
    wg_length=WG_LENGTH,
    wg_mode_padding=WG_MODE_PADDING,
    padding=PADDING,
    port_pml_offset=PORT_PML_OFFSET,
    variable_region_size=(1600 * u.nm, 1600 * u.nm),
    cladding_permittivity=CLADDING_PERMITTIVITY,
    slab_permittivity=SLAB_PERMITTIVITY,
    input_monitor_offset=INPUT_MONITOR_OFFSET,
    pml_width=PML_WIDTH_GRIDPOINTS,
)

WAVEGUIDE_BEND_MODEL = waveguide_bend.model.WaveguideBendModel(
    params=SIM_PARAMS, spec=_WAVEGUIDE_BEND_SPEC
)
LIGHTWEIGHT_WAVEGUIDE_BEND_MODEL = waveguide_bend.model.WaveguideBendModel(
    params=LIGHTWEIGHT_SIM_PARAMS, spec=_WAVEGUIDE_BEND_SPEC
)
WAVEGUIDE_BEND_TRANSMISSION_LOWER_BOUND = jnp.asarray(
    [
        0.0,  # |S11|^2 lower bound
        _linear_from_decibels(-0.5),  # |S21|^2 lower bound
    ]
)
WAVEGUIDE_BEND_TRANSMISSION_UPPER_BOUND = jnp.asarray(
    [
        _linear_from_decibels(-20.0),  # |S11|^2 upper bound
        1.0,  # |S21|^2 upper bound
    ]
)
WAVEGUIDE_BEND_SYMMETRIES = (symmetry.REFLECTION_NW_SE,)


# -----------------------------------------------------------------------------
# Wavelength demultiplexer with 3.2 x 3.2 or 6.4 x 6.4 um design region.
# -----------------------------------------------------------------------------


def _make_wdm_spec(
    design_extent_ij: u.Array,
    intended_sim_resolution: u.Quantity,
) -> wdm.spec.WdmSpec:
    """Construct a `wdm.spec.WdmSpec` with the specified design region extent.

    Since the `wdm.spec.WdmSpec` does not automatically compensate the simulation
    extent to account for changes in the PML region thickness with changes in the
    simulation resolution, the intended simulation resolution must be specified.
    The actual simulation extent appropriate for that resolution (i.e. accounting
    for the thickness of PML layers) is then computed.

    Note that if the actual simulation resolution is coarser than the intended
    resolution specified here, the PML may overlap with the source and monitors.
    Care should be taken to ensure the correct resolution is specified here.

    Args:
        design_extent_ij: Specifies the size of the design region.
        intended_sim_resolution: Specifies the simulation resolution to be used.

    Returns:
        The `wdm.spec.WdmSpec`.
    """

    design_extent_i, design_extent_j = design_extent_ij
    design_extent_i_nm: int = int(u.resolve(design_extent_i, 1 * u.nm))
    design_extent_j_nm: int = int(u.resolve(design_extent_j, 1 * u.nm))

    wg_width_nm: int = int(u.resolve(WG_WIDTH, 1 * u.nm))
    wg_length_nm: int = int(u.resolve(WG_LENGTH, 1 * u.nm))

    pad_width_nm: int = int(u.resolve(PADDING, 1 * u.nm))
    pml_width_nm: int = int(
        u.resolve(PML_WIDTH_GRIDPOINTS * intended_sim_resolution, 1 * u.nm)
    )

    extent_i_nm: int = int(design_extent_i_nm + 2 * wg_length_nm + 2 * pml_width_nm)
    extent_j_nm: int = int(design_extent_j_nm + 2 * pad_width_nm + 2 * pml_width_nm)

    return wdm.spec.WdmSpec(
        extent_ij=u.Array([extent_i_nm, extent_j_nm], u.nm),
        input_wg_j=extent_j_nm / 2 * u.nm,
        output_wgs_j=u.Array(
            (
                extent_j_nm / 2 - design_extent_j_nm / 2 + wg_width_nm / 2 + 320,
                extent_j_nm / 2 + design_extent_j_nm / 2 - wg_width_nm / 2 - 320,
            ),
            u.nm,
        ),
        wg_width=WG_WIDTH,
        wg_mode_padding=WG_MODE_PADDING,
        input_mode_i=pml_width_nm * u.nm + PORT_PML_OFFSET,
        output_mode_i=(extent_i_nm - pml_width_nm) * u.nm - PORT_PML_OFFSET,
        variable_region=(
            u.Array(
                (
                    extent_i_nm / 2 - design_extent_i_nm / 2,
                    extent_j_nm / 2 - design_extent_j_nm / 2,
                ),
                u.nm,
            ),
            u.Array(
                (
                    extent_i_nm / 2 + design_extent_i_nm / 2,
                    extent_j_nm / 2 + design_extent_j_nm / 2,
                ),
                u.nm,
            ),
        ),
        cladding_permittivity=CLADDING_PERMITTIVITY,
        slab_permittivity=SLAB_PERMITTIVITY,
        input_monitor_offset=INPUT_MONITOR_OFFSET,
        pml_width=PML_WIDTH_GRIDPOINTS,
    )


WDM_MODEL = wdm.model.WdmModel(
    params=SIM_PARAMS,
    spec=_make_wdm_spec(
        design_extent_ij=u.Array([6400, 6400], u.nm),
        intended_sim_resolution=SIM_PARAMS.resolution,
    ),
)
LIGHTWEIGHT_WDM_MODEL = wdm.model.WdmModel(
    params=LIGHTWEIGHT_SIM_PARAMS,
    spec=_make_wdm_spec(
        design_extent_ij=u.Array([3200, 3200], u.nm),
        intended_sim_resolution=LIGHTWEIGHT_SIM_PARAMS.resolution,
    ),
)
WDM_TRANSMISSION_LOWER_BOUND = jnp.asarray(
    [
        [
            [  # First wavelength band (1270 nm)
                0.0,  # |S11|^2 lower bound
                _linear_from_decibels(-3.0),  # |S21|^2 lower bound
                0.0,  # |S31|^2 lower bound
            ]
        ],
        [
            [  # Second wavelength band (1290 nm)
                0.0,  # |S11|^2 lower bound
                0.0,  # |S21|^2 lower bound
                _linear_from_decibels(-3.0),  # |S31|^2 lower bound
            ]
        ],
    ]
)
WDM_TRANSMISSION_UPPER_BOUND = jnp.asarray(
    [
        [
            [  # First wavelength band (1270 nm)
                _linear_from_decibels(-20.0),  # |S11|^2 upper bound
                1.0,  # |S21|^2 upper bound
                _linear_from_decibels(-20.0),  # |S31|^2 upper bound
            ]
        ],
        [
            [  # Second wavelength band (1290 nm)
                _linear_from_decibels(-20.0),  # |S11|^2 upper bound
                _linear_from_decibels(-20.0),  # |S21|^2 upper bound
                1.0,  # |S31|^2 upper bound
            ]
        ],
    ]
)



"""Defines functions common across diffract challenges.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
from typing import Any, Tuple, Union

import jax.numpy as jnp
import numpy as onp
from fmmax import basis, fields, fmm, scattering, utils  # type: ignore[import-untyped]
from jax import tree_util
from totypes import types

Params = Any

TE = "te"
TM = "tm"

DENSITY_LOWER_BOUND = 0.0
DENSITY_UPPER_BOUND = 1.0


@dataclasses.dataclass
class GratingSpec:
    """Defines the physical specifcation of a grating.

    Args:
        permittivity_ambient: Permittivity of the ambient material.
        permittivity_grating: Permittivity of the grating teeth.
        permittivity_encapsulation: Permittivity of the material in gaps between
            grating teeth.
        permittivity_substrate: Permittivity of the substrate.
        thickness_grating: Thickness of the grating layer.
        period_x: The size of the unit cell along the x direction.
        period_y: The size of the unit cell along the y direction.
    """

    permittivity_ambient: complex
    permittivity_grating: complex
    permittivity_encapsulation: complex
    permittivity_substrate: complex

    thickness_grating: float | jnp.ndarray

    period_x: float
    period_y: float


@dataclasses.dataclass
class GratingSimParams:
    """Parameters that configure the simulation of a grating.

    Attributes:
        grid_shape: The shape of the grid on which the permittivity is defined.
        wavelength: The wavelength of the excitation.
        polarization: The polarization of the excitation, TE or TM.
        formulation: The FMM formulation to be used.
        approximate_num_terms: Defines the number of terms in the Fourier expansion.
        truncation: Determines how the Fourier basis is truncated.
    """

    grid_shape: Tuple[int, int]
    wavelength: Union[float, jnp.ndarray]
    polarization: str
    formulation: fmm.Formulation
    approximate_num_terms: int
    truncation: basis.Truncation


@dataclasses.dataclass
class GratingResponse:
    """Contains the response of the grating.

    Attributes:
        wavelength: The wavelength for the efficiency calculation.
        transmission_efficiency: The per-order and per-wavelength coupling efficiency
            with which the excitation is transmitted.
        transmission_efficiency: The per-order and per-wavelength coupling efficiency
            with which the excitation is reflected.
        expansion: Defines the Fourier expansion for the calculation.
    """

    wavelength: jnp.ndarray
    transmission_efficiency: jnp.ndarray
    reflection_efficiency: jnp.ndarray
    expansion: basis.Expansion


tree_util.register_pytree_node(
    GratingResponse,
    lambda r: (
        (
            r.wavelength,
            r.transmission_efficiency,
            r.reflection_efficiency,
            r.expansion,
        ),
        None,
    ),
    lambda _, children: GratingResponse(*children),
)


def seed_density(grid_shape: Tuple[int, int], **kwargs: Any) -> types.Density2DArray:
    """Return the seed density for a grating component.

    Args:
        grid_shape: The shape of the grid on which the density is defined.
        kwargs: keyword arguments specifying additional properties of the seed
            density, e.g. symmetries.

    Returns:
        The seed density.
    """

    # Check kwargs that are required for a grating component.
    invalid_kwargs = ("array", "lower_bound", "upper_bound", "periodic")
    if any(k in invalid_kwargs for k in kwargs):
        raise ValueError(
            f"Attributes were specified which confict with automatically-extracted "
            f"attributes. Got {kwargs.keys()} when {invalid_kwargs} are automatically "
            f"extracted."
        )

    mid_density_value = (DENSITY_LOWER_BOUND + DENSITY_UPPER_BOUND) / 2
    return types.Density2DArray(
        array=jnp.full(grid_shape, mid_density_value),
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
        periodic=(True, True),
        **kwargs,
    )


def index_for_order(
    order: Tuple[int, int],
    expansion: basis.Expansion,
) -> int:
    """Returns the index for the specified Fourier order and expansion."""
    ((order_idx,),) = onp.where(onp.all(expansion.basis_coefficients == order, axis=1))
    assert tuple(expansion.basis_coefficients[order_idx, :]) == order
    return int(order_idx)


def grating_efficiency(
    density_array: jnp.ndarray,
    spec: GratingSpec,
    wavelength: jnp.ndarray,
    polarization: str,
    expansion: basis.Expansion,
    formulation: fmm.Formulation,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute the per-order transmission and reflection efficiency for a grating.

    The excitation for the calculation is a TE- or TM-polarized plane wave at the
    specified wavelength(s), incident from the substrate.

    Args:
        density_array: Defines the pattern of the grating layer.
        spec: Defines the physical specifcation of the grating.
        wavelength: The wavelength of the excitation.
        polarization: The polarization of the excitation, TE or TM.
        expansion: Defines the Fourier expansion for the calculation.
        formulation: Defines the FMM formulation to be used.

    Returns:
        The per-order transmission and reflection efficiency, having shape
        `(num_wavelengths, nkx, nky)`.
    """
    if polarization not in (TE, TM):
        raise ValueError(
            f"`polarization` must be one of {(TE, TM)} but got {polarization}."
        )

    permittivities = (
        jnp.full((1, 1), spec.permittivity_ambient),
        utils.interpolate_permittivity(
            permittivity_solid=jnp.asarray(spec.permittivity_grating),
            permittivity_void=jnp.asarray(spec.permittivity_encapsulation),
            density=density_array,
        ),
        jnp.full((1, 1), spec.permittivity_substrate),
    )

    layer_solve_results = [
        fmm.eigensolve_isotropic_media(
            wavelength=jnp.asarray(wavelength),
            in_plane_wavevector=jnp.zeros((2,)),  # normal incidence
            primitive_lattice_vectors=basis.LatticeVectors(
                u=spec.period_x * basis.X,
                v=spec.period_y * basis.Y,
            ),
            permittivity=p,
            expansion=expansion,
            formulation=formulation,
        )
        for p in permittivities
    ]

    # Layer thicknesses for the ambient and substrate are set to zero; these do not
    # affect the result of the calculation.
    layer_thicknesses = (
        jnp.zeros(()),
        jnp.asarray(spec.thickness_grating),
        jnp.zeros(()),
    )

    s_matrix = scattering.stack_s_matrix(layer_solve_results, layer_thicknesses)

    n = expansion.num_terms
    assert tuple(expansion.basis_coefficients[0, :]) == (0, 0)
    assert expansion.basis_coefficients.shape[0] == n

    # Generate the wave amplitudes for backward-going TE or TM-polarized plane waves
    # at the end of substrate layer.
    bwd_amplitude_silica_end = jnp.zeros((2 * n, 1), dtype=complex)
    if polarization == TE:
        bwd_amplitude_silica_end = bwd_amplitude_silica_end.at[0, 0].set(1.0)
    else:
        bwd_amplitude_silica_end = bwd_amplitude_silica_end.at[n, 0].set(1.0)

    # Calculate the incident power in the silca. Since the substrate thickness has
    # been set to zero, the forward and backward amplitudes are already colocated.
    fwd_amplitude_silica_start = s_matrix.s12 @ bwd_amplitude_silica_end
    fwd_flux_silica, bwd_flux_silica = fields.amplitude_poynting_flux(
        forward_amplitude=fwd_amplitude_silica_start,
        backward_amplitude=bwd_amplitude_silica_end,
        layer_solve_result=layer_solve_results[-1],
    )

    # Sum over orders and polarizations to get the total incident flux.
    total_incident_flux = jnp.sum(bwd_flux_silica, axis=-2, keepdims=True)

    # Calculate the transmitted power in the ambient.
    bwd_amplitude_ambient_end = s_matrix.s22 @ bwd_amplitude_silica_end
    _, bwd_flux_ambient = fields.amplitude_poynting_flux(
        forward_amplitude=jnp.zeros_like(bwd_amplitude_ambient_end),
        backward_amplitude=bwd_amplitude_ambient_end,
        layer_solve_result=layer_solve_results[0],
    )

    # Sum the fluxes over the two polarizations for each order.
    bwd_flux_ambient = bwd_flux_ambient[..., :n, :] + bwd_flux_ambient[..., n:, :]
    fwd_flux_silica = fwd_flux_silica[..., :n, :] + fwd_flux_silica[..., n:, :]

    transmission_efficiency = bwd_flux_ambient / total_incident_flux
    reflection_efficiency = fwd_flux_silica / total_incident_flux

    return transmission_efficiency, reflection_efficiency



"""Defines the ceviche challenges.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import functools
from typing import Any, Optional, Sequence, Tuple

import agjax  # type: ignore[import-untyped]
import jax
import jax.numpy as jnp
import numpy as onp
from jax import tree_util
from totypes import types

from invrs_gym.challenges import base
from invrs_gym.challenges.ceviche import defaults, transmission_loss
from invrs_gym.utils import initializers

Params = Any

FIELDS = "fields"
SPARAMS = "sparams"

TRANSMISSION_EXPONENT = 0.5
SCALAR_EXPONENT = 2.0

CEVICHE_DENSITY_LOWER_BOUND = 0.0
CEVICHE_DENSITY_UPPER_BOUND = 1.0


density_initializer = functools.partial(
    initializers.noisy_density_initializer,
    relative_mean=0.5,
    relative_noise_amplitude=0.1,
)


class CevicheComponent(base.Component):
    """Defines a general ceviche component."""

    def __init__(
        self,
        ceviche_model: defaults.Model,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs: Any,
    ) -> None:
        """Initialize a `CevicheComponent`.

        Args:
            ceviche_model: The model for the component.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """
        self.ceviche_model = ceviche_model
        self.seed_density = _seed_density(ceviche_model, **seed_density_kwargs)
        self.density_initializer = density_initializer

    def init(self, key: jax.Array) -> types.Density2DArray:
        """Return the initial parameters for the component."""
        params = self.density_initializer(key, self.seed_density)
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: types.Density2DArray,
        *,
        excite_port_idxs: Sequence[int] = (0,),
        wavelengths_nm: Optional[jnp.ndarray] = None,
        max_parallelizm: Optional[int] = None,
    ) -> Tuple[jnp.ndarray, base.AuxDict]:
        """Compute the response of the component and auxilliary quantities."""

        # The ceviche simulation function is autograd-differentiable. Wrap it so that
        # it can be differentiated using jax.

        @functools.partial(
            agjax.wrap_for_jax,
            nondiff_argnums=(1, 2, 3),
            nondiff_outputnums=(1,),
        )
        def sim_fn(
            design_variable: jnp.ndarray,
            excite_port_idxs: Sequence[int],
            wavelengths_nm: Optional[jnp.ndarray],
            max_parallelizm: Optional[int],
        ) -> Tuple[jnp.ndarray, onp.ndarray[Any, Any]]:
            s_params, fields = self.ceviche_model.simulate(
                design_variable, excite_port_idxs, wavelengths_nm, max_parallelizm
            )
            return s_params, fields

        sparams, fields = sim_fn(
            params.array, excite_port_idxs, wavelengths_nm, max_parallelizm
        )
        return sparams, {SPARAMS: sparams, FIELDS: fields}


def _seed_density(ceviche_model: defaults.Model, **kwargs: Any) -> types.Density2DArray:
    """Return the seed density for the `ceviche_model`.

    The seed density has shape and fixed pixels as required by the `ceviche_model`,
    and with other properties determined by keyword arguments.

    Args:
        ceviche_model: The model from which the seed density determined.
        kwargs: keyword arguments specifying additional properties of the seed
            density, e.g. symmetries.

    Returns:
        The seed density.
    """

    # Check kwargs that are computed from `ceviche_model`, which must not be specified.
    invalid_kwargs = (
        "array",
        "fixed_solid",
        "fixed_void",
        "lower_bound",
        "upper_bound",
    )
    if any(k in invalid_kwargs for k in kwargs):
        raise ValueError(
            f"Attributes were specified which confict with automatically-extracted "
            f"attributes. Got {kwargs.keys()} when {invalid_kwargs} are automatically "
            f"extracted."
        )

    shape = ceviche_model.design_variable_shape
    fixed_solid, fixed_void = _fixed_pixels(ceviche_model)
    mid_density_value = (CEVICHE_DENSITY_LOWER_BOUND + CEVICHE_DENSITY_UPPER_BOUND) / 2
    seed_density = types.Density2DArray(
        array=jnp.full(shape, mid_density_value),
        fixed_solid=fixed_solid,
        fixed_void=fixed_void,
        # For the ceviche challenges, density must lie between 0 and 1.
        lower_bound=CEVICHE_DENSITY_LOWER_BOUND,
        upper_bound=CEVICHE_DENSITY_UPPER_BOUND,
        **kwargs,
    )
    return seed_density


def _fixed_pixels(ceviche_model: defaults.Model) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Determine the fixed pixels for the given ceviche model.

    The fixed pixels exist around the border of the design, and are solid where the
    adjacent pixels outside the design are solid. They are void where the adjacent
    pixels outside the design are void.

    Args:
        ceviche_model: The model from which the fixed pixels are determined.

    Returns:
        The fixed solid and void pixels.
    """
    design_shape = ceviche_model.design_variable_shape
    density_fn = ceviche_model.density

    assert not onp.any(density_fn(jnp.ones(design_shape)) == 0.5)
    density = density_fn(jnp.full(design_shape, 0.5))

    i, j = onp.where(density == 0.5)
    i_lo = i[0]
    i_hi = i[-1]
    j_lo = j[0]
    j_hi = j[-1]
    assert i_lo > 0 and j_lo > 0
    assert i_hi < density.shape[0] - 1 and j_hi < density.shape[1] - 1

    # Compute the fixed solid pixels.
    fixed_solid = onp.zeros(design_shape, dtype=bool)
    fixed_solid[0, :] = density[i_lo - 1, j_lo : j_hi + 1] == 1
    fixed_solid[-1, :] = density[i_hi + 1, j_lo : j_hi + 1] == 1
    fixed_solid[:, 0] = density[i_lo : i_hi + 1, j_lo - 1] == 1
    fixed_solid[:, -1] = density[i_lo : i_hi + 1, j_hi + 1] == 1

    # Compute the fixed void pixels.
    fixed_void = onp.zeros(design_shape, dtype=bool)
    fixed_void[0, :] = density[i_lo - 1, j_lo : j_hi + 1] == 0
    fixed_void[-1, :] = density[i_hi + 1, j_lo : j_hi + 1] == 0
    fixed_void[:, 0] = density[i_lo : i_hi + 1, j_lo - 1] == 0
    fixed_void[:, -1] = density[i_lo : i_hi + 1, j_hi + 1] == 0

    return jnp.asarray(fixed_solid), jnp.asarray(fixed_void)


# -----------------------------------------------------------------------------
# Base class for ceviche challenges.
# -----------------------------------------------------------------------------


@dataclasses.dataclass
class CevicheChallenge(base.Challenge):
    """Defines a general ceviche challenge.

    The objective of the ceviche challenge is to find a component that whose
    transmission into its various ports lies within the target window defined
    by the transmission lower and upper bounds.

    Attributes:
        component: The component to be designed.
        transmission_lower_bound: The lower bound of the transmission window.
        transmission_upper_bound: The upper bound of the transmission window.
    """

    component: CevicheComponent
    transmission_lower_bound: jnp.ndarray
    transmission_upper_bound: jnp.ndarray

    def loss(self, response: jnp.ndarray) -> jnp.ndarray:
        """Compute a scalar loss from the component `response`."""
        # Power transmission is the squared magnitude of the scattering parameters.
        transmission = jnp.abs(response) ** 2
        # Repeat the bounds to match the transmission shape. Each value in the bounds
        # is then interpreted as the bounds for a wavelength band.
        lb = _wavelength_bound(self.transmission_lower_bound, transmission.shape)
        ub = _wavelength_bound(self.transmission_upper_bound, transmission.shape)
        return transmission_loss.orthotope_smooth_transmission_loss(
            transmission=transmission,
            window_lower_bound=lb,
            window_upper_bound=ub,
            transmission_exponent=jnp.asarray(TRANSMISSION_EXPONENT),
            scalar_exponent=jnp.asarray(SCALAR_EXPONENT),
        )

    def distance_to_target(self, response: jnp.ndarray) -> jnp.ndarray:
        """Compute distance from the component `response` to the challenge target."""
        transmission = jnp.abs(response) ** 2
        lb = _wavelength_bound(self.transmission_lower_bound, transmission.shape)
        ub = _wavelength_bound(self.transmission_upper_bound, transmission.shape)
        distance = transmission_loss.distance_to_window(
            transmission=transmission,
            window_lower_bound=lb,
            window_upper_bound=ub,
        )
        return distance

    def metrics(
        self,
        response: jnp.ndarray,
        params: types.Density2DArray,
        aux: base.AuxDict,
    ) -> base.AuxDict:
        """Compute challenge metrics."""
        del response, params, aux
        return {}


def _wavelength_bound(
    band_bound: jnp.ndarray, transmission_shape: Tuple[int, ...]
) -> jnp.ndarray:
    """Obtain per-wavelength bound compatible with `transmission_shape`."""
    assert len(transmission_shape) == 3
    assert band_bound.ndim <= 3

    # Add leading batch dimensions as needed.
    dims_to_add = range(0, len(transmission_shape) - band_bound.ndim)
    band_bound = jnp.expand_dims(band_bound, axis=dims_to_add)

    if not (transmission_shape[0] % band_bound.shape[0]) == 0:
        raise ValueError(
            f"Could not repeat `band_bound` with shape {band_bound.shape} to match "
            f"`transmission_shape` of {transmission_shape}; leading dimension size "
            f"must evenly divide the transmission shape."
        )

    repeats = transmission_shape[0] // band_bound.shape[0]
    repeated: jnp.ndarray = jnp.repeat(band_bound, repeats, axis=0)
    return repeated


# -----------------------------------------------------------------------------
# Constructors for ceviche challenges, including lightweight versions.
# -----------------------------------------------------------------------------


def beam_splitter(
    minimum_width: int = defaults.MINIMUM_WIDTH,
    minimum_spacing: int = defaults.MINIMUM_SPACING,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Beamsplitter with 3.2 x 2.0 um design and standard simulation params."""
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=defaults.BEAM_SPLITTER_MODEL,
            symmetries=defaults.BEAM_SPLITTER_SYMMETRIES,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.BEAM_SPLITTER_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.BEAM_SPLITTER_TRANSMISSION_UPPER_BOUND,
    )


def lightweight_beam_splitter(
    minimum_width: int = defaults.LIGHTWEIGHT_MINIMUM_WIDTH,
    minimum_spacing: int = defaults.LIGHTWEIGHT_MINIMUM_SPACING,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Beamsplitter with 3.2 x 2.0 um design and lightweight simulation params."""
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=defaults.LIGHTWEIGHT_BEAM_SPLITTER_MODEL,
            symmetries=defaults.BEAM_SPLITTER_SYMMETRIES,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.BEAM_SPLITTER_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.BEAM_SPLITTER_TRANSMISSION_UPPER_BOUND,
    )


def mode_converter(
    minimum_width: int = defaults.MINIMUM_WIDTH,
    minimum_spacing: int = defaults.MINIMUM_SPACING,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Mode converter with 1.6 x 1.6 um design and standard simulation params."""
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=defaults.MODE_CONVERTER_MODEL,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.MODE_CONVERTER_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.MODE_CONVERTER_TRANSMISSION_UPPER_BOUND,
    )


def lightweight_mode_converter(
    minimum_width: int = defaults.LIGHTWEIGHT_MINIMUM_WIDTH,
    minimum_spacing: int = defaults.LIGHTWEIGHT_MINIMUM_SPACING,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Mode converter with 1.6 x 1.6 um design and lightweight simulation params."""
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=defaults.LIGHTWEIGHT_MODE_CONVERTER_MODEL,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.MODE_CONVERTER_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.MODE_CONVERTER_TRANSMISSION_UPPER_BOUND,
    )


def waveguide_bend(
    minimum_width: int = defaults.MINIMUM_WIDTH,
    minimum_spacing: int = defaults.MINIMUM_SPACING,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Waveguide bend with 1.6 x 1.6 um design and standard simulation params."""
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=defaults.WAVEGUIDE_BEND_MODEL,
            symmetries=defaults.WAVEGUIDE_BEND_SYMMETRIES,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.WAVEGUIDE_BEND_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.WAVEGUIDE_BEND_TRANSMISSION_UPPER_BOUND,
    )


def lightweight_waveguide_bend(
    minimum_width: int = defaults.LIGHTWEIGHT_MINIMUM_WIDTH,
    minimum_spacing: int = defaults.LIGHTWEIGHT_MINIMUM_SPACING,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Waveguide bend with 1.6 x 1.6 um design and lightweight simulation params."""
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=defaults.LIGHTWEIGHT_WAVEGUIDE_BEND_MODEL,
            symmetries=defaults.WAVEGUIDE_BEND_SYMMETRIES,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.WAVEGUIDE_BEND_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.WAVEGUIDE_BEND_TRANSMISSION_UPPER_BOUND,
    )


def wdm(
    minimum_width: int = defaults.MINIMUM_WIDTH,
    minimum_spacing: int = defaults.MINIMUM_SPACING,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Demultiplexer with 6.4 x 6.4 um design and standard simulation params."""
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=defaults.WDM_MODEL,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.WDM_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.WDM_TRANSMISSION_UPPER_BOUND,
    )


def lightweight_wdm(
    minimum_width: int = defaults.LIGHTWEIGHT_MINIMUM_WIDTH,
    minimum_spacing: int = defaults.LIGHTWEIGHT_MINIMUM_SPACING,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Waveguide bend with 3.2 x 3.2 um design and lightweight simulation params."""
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=defaults.LIGHTWEIGHT_WDM_MODEL,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.WDM_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.WDM_TRANSMISSION_UPPER_BOUND,
    )



"""Defines the photon extractor challenge.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import functools
from typing import Tuple

import jax
from fmmax import basis, fmm  # type: ignore[import-untyped]
from jax import numpy as jnp
from jax import tree_util
from totypes import symmetry, types

from invrs_gym.challenges import base
from invrs_gym.challenges.extractor import component as extractor_component
from invrs_gym.utils import initializers

ENHANCEMENT_FLUX = "enhancement_flux"
ENHANCEMENT_FLUX_MEAN = "enhancement_flux_mean"
ENHANCEMENT_DOS = "enhancement_dos"
ENHANCEMENT_DOS_MEAN = "enhancement_dos_mean"


density_initializer = functools.partial(
    initializers.noisy_density_initializer,
    relative_mean=0.5,
    relative_noise_amplitude=0.1,
)


@dataclasses.dataclass
class PhotonExtractorChallenge(base.Challenge):
    """Defines the photon extractor challenge.

    The challenge is based on "Inverse-designed photon extractors for optically
    addressable defect qubits" by Chakravarthi et al. It involves optimizing a GaP
    patterned layer on diamond substrate above an implanted nitrogen vacancy defect.
    An oxide hard mask used to pattern the GaP is left in place after the etch.

    The goal of the optimization is to maximize extraction of 637 nm emission, i.e.
    to maximize the power coupled from the defect to the ambient above the extractor.

    https://opg.optica.org/optica/fulltext.cfm?uri=optica-7-12-1805

    Attributes:
        component: The component to be designed.
        bare_substratee_emitted_power: The power emitted by a nitrogen vacancy defect
            in a bare diamond substrate, i.e. without the GaP extractor structure.
        bare_substrate_collected_power: The power collected from a nitrogen vacancy
            defect in a bare diamond structure.
        flux_enhancement_lower_bound: Scalar giving the minimum target for flux
            enhancement. When the flux enhancement exceeds the lower bound, the
            challenge is considered solved.
    """

    component: extractor_component.ExtractorComponent
    bare_substrate_emitted_power: jnp.ndarray
    bare_substrate_collected_power: jnp.ndarray
    flux_enhancement_lower_bound: float

    def loss(self, response: extractor_component.ExtractorResponse) -> jnp.ndarray:
        """Compute a scalar loss from the component `response`."""
        # The response should have a length-3 trailing axis, corresponding to x, y,
        # and z-oriented dipoles.
        assert response.collected_power.shape[-1] == 3
        return -jnp.mean(response.collected_power)

    def distance_to_target(
        self, response: extractor_component.ExtractorResponse
    ) -> jnp.ndarray:
        """Compute distance from the component `response` to the challenge target."""
        enhancement_flux = (
            response.collected_power / self.bare_substrate_collected_power
        )
        return jnp.maximum(
            self.flux_enhancement_lower_bound - jnp.mean(enhancement_flux), 0.0
        )

    def metrics(
        self,
        response: extractor_component.ExtractorResponse,
        params: types.Density2DArray,
        aux: base.AuxDict,
    ) -> base.AuxDict:
        """Compute challenge metrics.

        Args:
            response: The response of the extractor component.
            params: The parameters where the response was evaluated.
            aux: The auxilliary quantities returned by the component response method.

        Returns:
            The metrics dictionary, with the following quantities:
                - mean enhancement of collected flux
                - mean enhancement of dipole density of states
                - the distance to the target flux enhancement
        """
        del params, aux
        enhancement_flux = (
            response.collected_power / self.bare_substrate_collected_power
        )
        enhancement_dos = response.emitted_power / self.bare_substrate_emitted_power
        return {
            ENHANCEMENT_FLUX: enhancement_flux,
            ENHANCEMENT_FLUX_MEAN: jnp.mean(enhancement_flux),
            ENHANCEMENT_DOS: enhancement_dos,
            ENHANCEMENT_DOS_MEAN: jnp.mean(enhancement_dos),
        }


EXTRACTOR_SPEC = extractor_component.ExtractorSpec(
    permittivity_ambient=(1.0 + 0.0j) ** 2,
    permittivity_oxide=(1.46 + 0.0j) ** 2,
    permittivity_extractor=(3.31 + 0.0j) ** 2,
    permittivity_substrate=(2.4102 + 0.0j) ** 2,
    thickness_ambient=1.0,
    thickness_oxide=0.13,
    thickness_extractor=0.25,
    thickness_substrate_before_source=0.1,
    thickness_substrate_after_source=0.9,
    width_design_region=1.5,
    width_padding=0.25,
    width_pml=0.4,
    fwhm_source=0.05,
    offset_monitor_source=0.025,
    offset_monitor_ambient=0.4,
    width_monitor_ambient=1.5,
)

EXTRACTOR_SIM_PARAMS = extractor_component.ExtractorSimParams(
    grid_spacing=0.01,
    wavelength=0.637,
    formulation=fmm.Formulation.JONES_DIRECT,
    approximate_num_terms=1200,
    truncation=basis.Truncation.CIRCULAR,
)

SYMMETRIES: Tuple[str, ...] = (
    symmetry.REFLECTION_N_S,
    symmetry.REFLECTION_E_W,
    symmetry.REFLECTION_NE_SW,
    symmetry.REFLECTION_NW_SE,
)

# Minimum width and spacing are 50 nm for the default dimensions.
MINIMUM_WIDTH = 5
MINIMUM_SPACING = 5

# Reference power values used to calculate the enhancement. These were computed
# by `compute_reference_response` with 1600 terms in the Fourier expansion.
BARE_SUBSTRATE_COLLECTED_POWER = jnp.asarray([2.469706, 2.469834, 0.13495])
BARE_SUBSTRATE_EMITTED_POWER = jnp.asarray([73.41745, 73.41583, 84.21051])

# Target is to achieve flux enhancement of 50 times or greater.
FLUX_ENHANCEMENT_LOWER_BOUND = 50.0


def photon_extractor(
    minimum_width: int = MINIMUM_WIDTH,
    minimum_spacing: int = MINIMUM_SPACING,
    density_initializer: base.DensityInitializer = density_initializer,
    bare_substrate_emitted_power: jnp.ndarray = BARE_SUBSTRATE_EMITTED_POWER,
    bare_substrate_collected_power: jnp.ndarray = BARE_SUBSTRATE_COLLECTED_POWER,
    flux_enhancement_lower_bound: float = FLUX_ENHANCEMENT_LOWER_BOUND,
    spec: extractor_component.ExtractorSpec = EXTRACTOR_SPEC,
    sim_params: extractor_component.ExtractorSimParams = EXTRACTOR_SIM_PARAMS,
    symmetries: Tuple[str, ...] = SYMMETRIES,
) -> PhotonExtractorChallenge:
    """Photon extractor with 1.5 x 1.5 um design region.

    The challenge is based on "Inverse-designed photon extractors for optically
    addressable defect qubits" by Chakravarthi et al. It involves optimizing a GaP
    patterned layer on diamond substrate above an implanted nitrogen vacancy defect.
    An oxide hard mask used to pattern the GaP is left in place after the etch.

    The goal of the optimization is to maximize extraction of 637 nm emission, i.e.
    to maximize the power coupled from the defect to the ambient above the extractor.
    https://opg.optica.org/optica/fulltext.cfm?uri=optica-7-12-1805

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 180 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        density_initializer: Callble which returns the initial density, given a
            key and seed density.
        bare_substrate_emitted_power: The power emitted by a nitrogen vacancy defect
            in a bare diamond substrate, i.e. without the GaP extractor structure.
        bare_substrate_collected_power: The power collected from a nitrogen vacancy
            defect in a bare diamond structure.
        flux_enhancement_lower_bound: Scalar giving the minimum target for flux
            enhancement. When the flux enhancement exceeds the lower bound, the
            challenge is considered solved.
        spec: Defines the physical specification of the metagrating.
        sim_params: Defines the simulation settings of the metagrating.
        symmetries: Defines the symmetries of the metagrating.

    Returns:
        The `PhotonExtractorChallenge`.
    """
    return PhotonExtractorChallenge(
        component=extractor_component.ExtractorComponent(
            spec=spec,
            sim_params=sim_params,
            density_initializer=density_initializer,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            symmetries=symmetries,
        ),
        bare_substrate_emitted_power=bare_substrate_emitted_power,
        bare_substrate_collected_power=bare_substrate_collected_power,
        flux_enhancement_lower_bound=flux_enhancement_lower_bound,
    )


def bare_substrate_response(
    spec: extractor_component.ExtractorSpec = EXTRACTOR_SPEC,
    sim_params: extractor_component.ExtractorSimParams = EXTRACTOR_SIM_PARAMS,
) -> extractor_component.ExtractorResponse:
    """Computes the response of the nitrogen vacancy in a bare diamond substrate."""
    component = extractor_component.ExtractorComponent(
        spec=spec,
        sim_params=sim_params,
        density_initializer=lambda _, d: tree_util.tree_map(jnp.zeros_like, d),
    )
    params = component.init(jax.random.PRNGKey(0))
    response, _ = component.response(params)
    return response



"""Base types for challenges and components.

Copyright (c) 2023 The INVRS-IO authors.
"""

import abc
from typing import Any, Callable, Dict, Tuple

import jax
import jax.numpy as jnp
from totypes import types

AuxDict = Dict[str, Any]
PyTree = Any


DensityInitializer = Callable[[jax.Array, types.Density2DArray], types.Density2DArray]


class Component(abc.ABC):
    """Base class for components.

    The component represents the physical structure to be optimized. Initial parameters
    are returned by the `init` method; these can include types having metadata that
    specifies bounds on parameter values or target characteristics, e.g. a minimum
    feature size.

    The component has a default "excitation", e.g. illumination with a particular
    wavelength of light from a specific direction; the response of the component to the
    excitation is obtained with the `response` method. The `response` method of classes
    inheriting from `Component` will generally have optional arguments to modify the
    excitation condition.
    """

    @abc.abstractmethod
    def init(self, key: jax.Array):
        """Returns the initial parameters given a `PRNGKey`."""

    @abc.abstractmethod
    def response(self, params: PyTree) -> Tuple[Any, AuxDict]:
        """Returns the response of the component for the given `params`."""


class Challenge(abc.ABC):
    """Base class for challenges.

    Challenges consist of a `Component` to be optimized and a loss function which
    returns a scalar loss given the component response.

    Challenges also include a `distance_to_target` function which returns a scalar
    that is zero or negative when the challenge objective is achieved. Unlike the
    loss function, `distance_to_target` may not be well-suited for gradient-based
    optimization and may not even be differentiable. The distance serves as an
    independent way to assess the quality of a solution, e.g. in cases where various
    loss functions are compared.

    Challenges also include a `metrics` function which computes additional quantities
    that may be useful assessing the quality of a particular component.

    Attributes:
        component: The `Component` to be optimized.
    """

    component: Component

    @abc.abstractmethod
    def loss(self, response: Any) -> jnp.ndarray:
        """Compute scalar loss for the `response`."""

    @abc.abstractmethod
    def distance_to_target(self, response: Any) -> jnp.ndarray:
        """Compute the distance between the response and the challenge target."""

    @abc.abstractmethod
    def metrics(self, response: Any, params: PyTree, aux: AuxDict) -> AuxDict:
        """Compute metrics for a component response and associated quantities."""



"""Defines loss functions that target transmission values into various output ports.

Copyright (c) 2023 The INVRS-IO authors.
"""

import jax
import jax.numpy as jnp

_MIN_PHYSICAL_TRANSMISSION = 0.0
_MAX_PHYSICAL_TRANSMISSION = 1.0


def orthotope_smooth_transmission_loss(
    transmission: jnp.ndarray,
    window_lower_bound: jnp.ndarray,
    window_upper_bound: jnp.ndarray,
    transmission_exponent: jnp.ndarray,
    scalar_exponent: jnp.ndarray,
) -> jnp.ndarray:
    """Compute a scalar loss from a array based on an orthotope transmission window.

    The loss is related to an orthotope window in tansmission space, i.e. the space
    of the squared magnitude of scattering parameters. With values for
    `transmission_exponent` and `scalar_exponent` of `1.0` and `2.0`, respectively,
    this loss function is equivalent to that of [2022 Schubert]
    (https://arxiv.org/abs/2201.12965).

    Args:
        transmission: The transmission array for which the loss is to be calculated.
        window_lower_bound: Array giving the lower bound of the target transmission
            window. Must be broadcast compatible with `transmission`.
        window_upper_bound: Defines the upper bound of the target window.
        transmission_exponent: Exponent applied to the transmission and window
            bounds prior to scalarization.
        scalar_exponent: Exponent applied to the final scalar loss.

    Returns:
        The scalar loss value.
    """
    # Compute the signed psuedodistance. This is equal to the signed distance to the
    # nearest bound, except when the bounds are the min and max physical transmission
    # values, in which case the distance is equal to the window size.
    elementwise_signed_psuedodistance = elementwise_signed_psuedodistance_to_window(
        transmission=transmission**transmission_exponent,
        window_lower_bound=window_lower_bound**transmission_exponent,
        window_upper_bound=window_upper_bound**transmission_exponent,
    )

    # Scale the signed distance by the maximum window size.
    lower_bound = jnp.maximum(window_lower_bound, _MIN_PHYSICAL_TRANSMISSION)
    upper_bound = jnp.minimum(window_upper_bound, _MAX_PHYSICAL_TRANSMISSION)
    window_size = upper_bound - lower_bound
    elementwise_signed_psuedodistance /= jnp.amin(window_size)

    transformed_elementwise_signed_distance = jax.nn.softplus(
        elementwise_signed_psuedodistance
    )
    loss: jnp.ndarray = (
        jnp.linalg.norm(transformed_elementwise_signed_distance) ** scalar_exponent
    )
    return loss


def distance_to_window(
    transmission: jnp.ndarray,
    window_lower_bound: jnp.ndarray,
    window_upper_bound: jnp.ndarray,
) -> jnp.ndarray:
    """Returns the distance to a transmission window.

    When the transmission lies inside the transmission window, the distance to
    the window is zero.

    Args:
        transmission: The transmission for which the signed distance is sought.
        window_lower_bound: Array defining the transmission window lower bound.
        window_upper_bound: Array defining the transmission window upper bound.

    Returns:
        The elementwise signed distance.
    """
    elementwise_signed_distance = elementwise_signed_psuedodistance_to_window(
        transmission=transmission,
        window_lower_bound=window_lower_bound,
        window_upper_bound=window_upper_bound,
    )
    elementwise_distance = jnp.maximum(elementwise_signed_distance, 0.0)
    distance: jnp.ndarray = jnp.linalg.norm(elementwise_distance)
    return distance


def elementwise_signed_psuedodistance_to_window(
    transmission: jnp.ndarray,
    window_lower_bound: jnp.ndarray,
    window_upper_bound: jnp.ndarray,
) -> jnp.ndarray:
    """Returns the elementwise signed psuedodistance to a transmission window.

    The psuedodistance is given by,

        - the distance to the nearest bound, when both bounds are within the
          window (e.g. the lower bound is greater than the minimum physical
          transmission value).
        - the distance to the upper bound, when the lower bound is less than or
          equal to the minimum physical transmission value, and the upper bound
          is less than the maximum physical transmission value.
        - the distance to the lower bound, when the upper bound is greater than
          or equal to the maximum physicall transmission value, and the lower
          bound is greater than the minimum physical transmission value.
        - the difference between the maximum and minimum physical transmission
          value, when both bounds equal or exceed their physical extremal values.
          In this case, the psuedodistance has no dependence on `transmission`.

    Args:
        transmission: The transmission for which the signed distance is sought.
        window_lower_bound: Array defining the transmission window lower bound.
        window_upper_bound: Array defining the transmission window upper bound.

    Returns:
        The elementwise signed psuedodistance.
    """
    # Signed distance to lower bound is positive when `transmission` is below the lower
    # bound and negative when it is above the lower bound. When the lower bound is less
    # or equal to the minimum physical transmission, disregard the distance entirely
    # and substitute a value equal to the signed distance to the upper bound when the
    # response value is equal to the lower bound.
    elementwise_signed_distance_to_lower_bound = window_lower_bound - transmission
    elementwise_signed_distance_to_lower_bound = jnp.where(
        window_lower_bound > _MIN_PHYSICAL_TRANSMISSION,
        elementwise_signed_distance_to_lower_bound,
        -(_MAX_PHYSICAL_TRANSMISSION - _MIN_PHYSICAL_TRANSMISSION),
    )

    # Signed distance to upper bound defined similarly to the lower bound.
    elementwise_signed_distance_to_upper_bound = transmission - window_upper_bound
    elementwise_signed_distance_to_upper_bound = jnp.where(
        window_upper_bound < _MAX_PHYSICAL_TRANSMISSION,
        elementwise_signed_distance_to_upper_bound,
        -(_MAX_PHYSICAL_TRANSMISSION - _MIN_PHYSICAL_TRANSMISSION),
    )

    return jnp.maximum(
        elementwise_signed_distance_to_lower_bound,
        elementwise_signed_distance_to_upper_bound,
    )



"""invrs_gym - A collection of inverse design challenges.

Copyright (c) 2023 The INVRS-IO authors.
"""

__version__ = "v0.2.1"
__author__ = "Martin F. Schubert <mfschubert@gmail.com>"

from invrs_gym import challenges as challenges
from invrs_gym import utils as utils



"""Defines the diffractive beamsplitter challenge.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import functools
import itertools
from typing import Any, Callable, Dict, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp
from fmmax import basis, fmm  # type: ignore[import-untyped]
from jax import tree_util
from totypes import types

from invrs_gym.challenges import base
from invrs_gym.challenges.diffract import common
from invrs_gym.utils import initializers

Params = Dict[str, types.BoundedArray | types.Density2DArray]
ThicknessInitializer = Callable[[jax.Array, types.BoundedArray], types.BoundedArray]


DENSITY = "density"
THICKNESS = "thickness"

TOTAL_EFFICIENCY = "total_efficiency"
AVERAGE_EFFICIENCY = "average_efficiency"
MIN_EFFICIENCY = "min_efficiency"
ZEROTH_ORDER_EFFICIENCY = "zeroth_order_efficiency"
ZEROTH_ORDER_ERROR = "zeroth_order_error"
UNIFORMITY_ERROR = "uniformity_error"
UNIFORMITY_ERROR_WITHOUT_ZEROTH_ORDER = "uniformity_error_without_zeroth_order"

density_initializer = functools.partial(
    initializers.noisy_density_initializer,
    relative_mean=0.5,
    relative_noise_amplitude=0.1,
)


class DiffractiveSplitterComponent(base.Component):
    """Defines a diffractive splitter component."""

    def __init__(
        self,
        spec: common.GratingSpec,
        sim_params: common.GratingSimParams,
        thickness_initializer: ThicknessInitializer,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs: Any,
    ) -> None:
        """Initializes the grating component.

        Args:
            spec: Defines the physical specification of the splitter.
            sim_params: Defines simulation parameters for the splitter.
            thickness_initializer: Callable which returns the initial thickness for
                the grating layer from a random key and a bounded array with value
                equal the thickness from `spec`.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """

        self.spec = spec
        self.sim_params = sim_params
        self.thickness_initializer = thickness_initializer
        self.density_initializer = density_initializer

        self.seed_thickness = types.BoundedArray(
            array=self.spec.thickness_grating,
            lower_bound=0.0,
            upper_bound=None,
        )
        self.seed_density = common.seed_density(
            self.sim_params.grid_shape, **seed_density_kwargs
        )

        self.expansion = basis.generate_expansion(
            primitive_lattice_vectors=basis.LatticeVectors(
                u=self.spec.period_x * basis.X,
                v=self.spec.period_y * basis.Y,
            ),
            approximate_num_terms=self.sim_params.approximate_num_terms,
            truncation=self.sim_params.truncation,
        )

    def init(self, key: jax.Array) -> Params:
        """Return the initial parameters for the diffractive splitter component."""
        key_thickness, key_density = jax.random.split(key)
        params = {
            THICKNESS: self.thickness_initializer(key_thickness, self.seed_thickness),
            DENSITY: self.density_initializer(key_density, self.seed_density),
        }
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: Params,
        *,
        wavelength: Optional[Union[float, jnp.ndarray]] = None,
        expansion: Optional[basis.Expansion] = None,
    ) -> Tuple[common.GratingResponse, base.AuxDict]:
        """Computes the response of the diffractive splitter.

        Args:
            params: The parameters defining the diffractive splitter, matching those
                returned by the `init` method.
            wavelength: Optional wavelength to override the default in `sim_params`.
            expansion: Optional expansion to override the default `expansion`.

        Returns:
            The `(response, aux)` tuple.
        """
        if expansion is None:
            expansion = self.expansion
        if wavelength is None:
            wavelength = self.sim_params.wavelength
        spec = dataclasses.replace(
            self.spec,
            thickness_grating=jnp.asarray(params[THICKNESS].array),
        )
        transmission_efficiency, reflection_efficiency = common.grating_efficiency(
            density_array=params[DENSITY].array,  # type: ignore[arg-type]
            spec=spec,
            wavelength=jnp.asarray(wavelength),
            polarization=self.sim_params.polarization,
            expansion=expansion,
            formulation=self.sim_params.formulation,
        )
        response = common.GratingResponse(
            wavelength=jnp.asarray(wavelength),
            transmission_efficiency=transmission_efficiency,
            reflection_efficiency=reflection_efficiency,
            expansion=expansion,
        )
        return response, {}


@dataclasses.dataclass
class DiffractiveSplitterChallenge(base.Challenge):
    """Defines the diffractive beamsplitter challenge.

    The objective of the challenge is to evenly split light into an array of
    transmitted diffraction orders. The challenge is based on the LightTrans
    publication, "Design and rigorous analysis of a non-paraxial diffractive
    beamsplitter", retrieved from
    https://www.lighttrans.com/fileadmin/shared/UseCases/Application_UC_Rigorous%20Analysis%20of%20Non-paraxial%20Diffractive%20Beam%20Splitter.pdf

    The challenge is considered solved when the efficiency of each order is greater
    than `normalized_efficiency_lower_bound / num_splits` and less than
    `normalized_efficiency_upper_bound / num_splits`.

    Attributes:
        component: The component to be designed.
        splitting: Defines the target splitting for the beamsplitter.
        normalized_efficiency_lower_bound: The lower bound for normalized efficiency.
        normalized_efficiency_upper_bound: The upper bound for normalized efficiency.
    """

    component: DiffractiveSplitterComponent
    splitting: Tuple[int, int]
    normalized_efficiency_lower_bound: float
    normalized_efficiency_upper_bound: float

    def loss(self, response: common.GratingResponse) -> jnp.ndarray:
        """Compute a scalar loss from the component `response`."""
        assert response.transmission_efficiency.shape[-1] == 1

        efficiency = extract_orders_for_splitting(
            response.transmission_efficiency,
            expansion=response.expansion,
            splitting=self.splitting,
        )
        assert efficiency.shape[-3:] == self.splitting + (1,)
        num_splits = self.splitting[0] * self.splitting[1]
        return jnp.linalg.norm(jnp.sqrt(1 / num_splits) - jnp.sqrt(efficiency)) ** 2

    def distance_to_target(self, response: common.GratingResponse) -> jnp.ndarray:
        """Compute distance from the component `response` to the challenge target."""
        efficiency = extract_orders_for_splitting(
            response.transmission_efficiency,
            expansion=response.expansion,
            splitting=self.splitting,
        )
        assert efficiency.shape[-3:] == self.splitting + (1,)
        num_splits = self.splitting[0] * self.splitting[1]
        lower_bound = self.normalized_efficiency_lower_bound / num_splits
        upper_bound = self.normalized_efficiency_upper_bound / num_splits
        lower_bound_error = jnp.maximum(0, lower_bound - efficiency)
        upper_bound_error = jnp.maximum(0, efficiency - upper_bound)
        error = jnp.maximum(upper_bound_error, lower_bound_error)
        return jnp.linalg.norm(error)

    def metrics(
        self,
        response: common.GratingResponse,
        params: types.Density2DArray,
        aux: base.AuxDict,
    ) -> base.AuxDict:
        """Compute challenge metrics.

        Args:
            response: The response of the diffractive splitter component.
            params: The parameters where the response was evaluated.
            aux: The auxilliary quantities returned by the component response method.

        Returns:
            The metrics dictionary, with the following quantities:
                - total efficiency
                - average efficiency
                - zeroth order efficiency
                - zeroth order error
                - uniformity error
                - uniformity error without zeroth order
            For details, see slide 14 of the LightTrans publication.
        """
        del params, aux
        transmission = extract_orders_for_splitting(
            response.transmission_efficiency,
            expansion=response.expansion,
            splitting=self.splitting,
        )
        assert transmission.shape[-3:] == self.splitting + (1,)
        # Metrics are averaged over the wavelength axis, if one exists.
        total_efficiency = jnp.mean(jnp.sum(transmission, axis=(-3, -2, -1)))
        average_efficiency = jnp.mean(transmission)
        min_efficiency = jnp.amin(transmission)

        i0 = self.splitting[0] // 2
        j0 = self.splitting[1] // 2
        zeroth_efficiency = jnp.mean(transmission[..., i0, j0, :])
        zeroth_error = (zeroth_efficiency - average_efficiency) / average_efficiency

        uniformity_error = (jnp.amax(transmission) - jnp.amin(transmission)) / (
            jnp.amax(transmission) + jnp.amin(transmission)
        )

        mask = jnp.ones(transmission.shape, dtype=bool)
        mask = mask.at[..., i0, j0, :].set(False)
        masked_max_transmission = jnp.amax(jnp.where(mask, transmission, 0.0))
        masked_min_transmission = jnp.amin(jnp.where(mask, transmission, 1.0))

        uniformity_error_without_zeroth = (
            masked_max_transmission - masked_min_transmission
        ) / (masked_max_transmission + masked_min_transmission)

        return {
            TOTAL_EFFICIENCY: total_efficiency,
            AVERAGE_EFFICIENCY: average_efficiency,
            MIN_EFFICIENCY: min_efficiency,
            ZEROTH_ORDER_EFFICIENCY: zeroth_efficiency,
            ZEROTH_ORDER_ERROR: zeroth_error,
            UNIFORMITY_ERROR: uniformity_error,
            UNIFORMITY_ERROR_WITHOUT_ZEROTH_ORDER: uniformity_error_without_zeroth,
        }


def indices_for_splitting(
    expansion: basis.Expansion,
    splitting: Tuple[int, int],
) -> Tuple[int, ...]:
    """Return the indices for the orders identified by the target `splitting`."""
    idxs = []
    orders_x = range(-splitting[0] // 2 + 1, splitting[0] // 2 + 1)
    orders_y = range(-splitting[1] // 2 + 1, splitting[1] // 2 + 1)
    for ox, oy in itertools.product(orders_x, orders_y):
        idxs.append(common.index_for_order((ox, oy), expansion))
    return tuple(idxs)


def extract_orders_for_splitting(
    array: jnp.ndarray,
    expansion: basis.Expansion,
    splitting: Tuple[int, int],
) -> jnp.ndarray:
    """Extract the values from `array` for the specified splitting."""

    num_splits = splitting[0] * splitting[1]
    shape = array.shape[:-2] + splitting + (array.shape[-1],)
    flat_shape = array.shape[:-2] + (num_splits, array.shape[-1])

    idxs = indices_for_splitting(expansion, splitting)
    extracted = jnp.zeros(flat_shape, dtype=array.dtype)
    extracted = extracted.at[..., :, :].set(array[..., idxs, :])
    return extracted.reshape(shape)


# -----------------------------------------------------------------------------
# Diffractive splitter with 7.2 x 7.2 um design region.
# -----------------------------------------------------------------------------


DIFFRACTIVE_SPLITTER_SPEC = common.GratingSpec(
    permittivity_ambient=(1.46 + 0.0j) ** 2,
    # Small imaginary part stabilizes the FMM calculation.
    permittivity_grating=(1.46 + 0.00001j) ** 2,
    permittivity_encapsulation=(1.0 + 0.00001j) ** 2,
    permittivity_substrate=(1.0 + 0.0j) ** 2,
    thickness_grating=0.692,
    period_x=7.2,
    period_y=7.2,
)

DIFFRACTIVE_SPLITTER_SIM_PARAMS = common.GratingSimParams(
    grid_shape=(180, 180),
    wavelength=0.6328,
    polarization=common.TM,
    formulation=fmm.Formulation.JONES_DIRECT,
    approximate_num_terms=800,
    truncation=basis.Truncation.CIRCULAR,
)

# Objective is to split into a 7 x 7 array of beams. The minimum efficiency of any
# beam shhould be `0.7 / (7 * 7)`, while the maximum should be `1.3 / (7 * 7)`.
SPLITTING = (7, 7)
NORMALIZED_EFFICIENCY_LOWER_BOUND = 0.7
NORMALIZED_EFFICIENCY_UPPER_BOUND = 1.3


def diffractive_splitter(
    minimum_width: int = 10,
    minimum_spacing: int = 10,
    thickness_initializer: ThicknessInitializer = initializers.identity_initializer,
    density_initializer: base.DensityInitializer = density_initializer,
    splitting: Tuple[int, int] = SPLITTING,
    normalized_efficiency_lower_bound: float = NORMALIZED_EFFICIENCY_LOWER_BOUND,
    normalized_efficiency_upper_bound: float = NORMALIZED_EFFICIENCY_UPPER_BOUND,
    spec: common.GratingSpec = DIFFRACTIVE_SPLITTER_SPEC,
    sim_params: common.GratingSimParams = DIFFRACTIVE_SPLITTER_SIM_PARAMS,
    symmetries: Sequence[str] = (),
) -> DiffractiveSplitterChallenge:
    """Non-paraxial diffractive beamsplitter challenge.

    The diffractive splitter is based on "Design and rigorous analysis of a
    non-paraxial diffractive beamsplitter", an example of the LightTrans software
    (https://www.lighttrans.com/use-cases/application/design-and-rigorous-analysis-of-non-paraxial-diffractive-beam-splitter.html).

    It involves splitting a normally-incident TM-polarized plane wave into an
    array of 7x7 beams with maximal efficiency and uniformity. The challenge is
    considered solved when the efficiency of each order is greater than
    `normalized_efficiency_lower_bound / num_splits` and less than
    `normalized_efficiency_upper_bound / num_splits`.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 180 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        thickness_initializer: Callble which returns the initial thickness, given a
            key and seed thickness.
        density_initializer: Callble which returns the initial density, given a
            key and seed density.
        splitting: Defines shape of the beam array to be created by the splitter.
        normalized_efficiency_lower_bound: The lower bound for normalized efficiency.
        normalized_efficiency_upper_bound: The upper bound for normalized efficiency.
        spec: Defines the physical specification of the metagrating.
        sim_params: Defines the simulation settings of the metagrating.
        symmetries: Defines the symmetries of the metagrating.

    Returns:
        The `MetagratingChallenge`.
    """
    return DiffractiveSplitterChallenge(
        component=DiffractiveSplitterComponent(
            spec=spec,
            sim_params=sim_params,
            thickness_initializer=thickness_initializer,
            density_initializer=density_initializer,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            symmetries=symmetries,
        ),
        splitting=splitting,
        normalized_efficiency_lower_bound=normalized_efficiency_lower_bound,
        normalized_efficiency_upper_bound=normalized_efficiency_upper_bound,
    )
